@@ -5,80 +5,169 @@
 # This code is distributed under the terms and conditions
 # from the MIT License (MIT).
 #
-"""Implements file-like objects for reading and writing from/to S3."""
+"""Implements file-like objects for reading and writing from/to AWS S3."""
 
 import io
-import contextlib
 import functools
 import logging
 import time
-import warnings
 
-import boto3
-import botocore.client
-import botocore.exceptions
-import six
+try:
+    import boto3
+    import botocore.client
+    import botocore.exceptions
+except ImportError:
+    MISSING_DEPS = True
 
 import smart_open.bytebuffer
+import smart_open.concurrency
+import smart_open.utils
+
+from smart_open import constants
 
 logger = logging.getLogger(__name__)
 
-# AWS Lambda environments do not support multiprocessing.Queue or multiprocessing.Pool.
-# However they do support Threads and therefore concurrent.futures's ThreadPoolExecutor.
-# We use this flag to allow python 2 backward compatibility, where concurrent.futures doesn't exist.
-_CONCURRENT_FUTURES = False
-try:
-    import concurrent.futures
-    _CONCURRENT_FUTURES = True
-except ImportError:
-    warnings.warn("concurrent.futures could not be imported and won't be used")
-
-# Multiprocessing is unavailable in App Engine (and possibly other sandboxes).
-# The only method currently relying on it is iter_bucket, which is instructed
-# whether to use it by the MULTIPROCESSING flag.
-_MULTIPROCESSING = False
-try:
-    import multiprocessing.pool
-    _MULTIPROCESSING = True
-except ImportError:
-    warnings.warn("multiprocessing could not be imported and won't be used")
-
-
 DEFAULT_MIN_PART_SIZE = 50 * 1024**2
 """Default minimum part size for S3 multipart uploads"""
 MIN_MIN_PART_SIZE = 5 * 1024 ** 2
 """The absolute minimum permitted by Amazon."""
-READ_BINARY = 'rb'
-WRITE_BINARY = 'wb'
-MODES = (READ_BINARY, WRITE_BINARY)
-"""Allowed I/O modes for working with S3."""
 
-BINARY_NEWLINE = b'\n'
-
-SUPPORTED_SCHEMES = ("s3", "s3n", 's3u', "s3a")
+SCHEMES = ("s3", "s3n", 's3u', "s3a")
+DEFAULT_PORT = 443
+DEFAULT_HOST = 's3.amazonaws.com'
 
 DEFAULT_BUFFER_SIZE = 128 * 1024
 
-START = 0
-CURRENT = 1
-END = 2
-WHENCE_CHOICES = [START, CURRENT, END]
+URI_EXAMPLES = (
+    's3://my_bucket/my_key',
+    's3://my_key:my_secret@my_bucket/my_key',
+    's3://my_key:my_secret@my_server:my_port@my_bucket/my_key',
+)
 
 _UPLOAD_ATTEMPTS = 6
 _SLEEP_SECONDS = 10
 
-
-def clamp(value, minval, maxval):
-    return max(min(value, maxval), minval)
+# Returned by AWS when we try to seek beyond EOF.
+_OUT_OF_RANGE = 'Requested Range Not Satisfiable'
 
 
-def make_range_string(start, stop=None):
+def parse_uri(uri_as_string):
     #
-    # https://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.35
+    # Restrictions on bucket names and labels:
     #
-    if stop is None:
-        return 'bytes=%d-' % start
-    return 'bytes=%d-%d' % (start, stop)
+    # - Bucket names must be at least 3 and no more than 63 characters long.
+    # - Bucket names must be a series of one or more labels.
+    # - Adjacent labels are separated by a single period (.).
+    # - Bucket names can contain lowercase letters, numbers, and hyphens.
+    # - Each label must start and end with a lowercase letter or a number.
+    #
+    # We use the above as a guide only, and do not perform any validation.  We
+    # let boto3 take care of that for us.
+    #
+    split_uri = smart_open.utils.safe_urlsplit(uri_as_string)
+    assert split_uri.scheme in SCHEMES
+
+    port = DEFAULT_PORT
+    host = DEFAULT_HOST
+    ordinary_calling_format = False
+    #
+    # These defaults tell boto3 to look for credentials elsewhere
+    #
+    access_id, access_secret = None, None
+
+    #
+    # Common URI template [secret:key@][host[:port]@]bucket/object
+    #
+    # The urlparse function doesn't handle the above schema, so we have to do
+    # it ourselves.
+    #
+    uri = split_uri.netloc + split_uri.path
+
+    if '@' in uri and ':' in uri.split('@')[0]:
+        auth, uri = uri.split('@', 1)
+        access_id, access_secret = auth.split(':')
+
+    head, key_id = uri.split('/', 1)
+    if '@' in head and ':' in head:
+        ordinary_calling_format = True
+        host_port, bucket_id = head.split('@')
+        host, port = host_port.split(':', 1)
+        port = int(port)
+    elif '@' in head:
+        ordinary_calling_format = True
+        host, bucket_id = head.split('@')
+    else:
+        bucket_id = head
+
+    return dict(
+        scheme=split_uri.scheme,
+        bucket_id=bucket_id,
+        key_id=key_id,
+        port=port,
+        host=host,
+        ordinary_calling_format=ordinary_calling_format,
+        access_id=access_id,
+        access_secret=access_secret,
+    )
+
+
+def _consolidate_params(uri, transport_params):
+    """Consolidates the parsed Uri with the additional parameters.
+
+    This is necessary because the user can pass some of the parameters can in
+    two different ways:
+
+    1) Via the URI itself
+    2) Via the transport parameters
+
+    These are not mutually exclusive, but we have to pick one over the other
+    in a sensible way in order to proceed.
+
+    """
+    transport_params = dict(transport_params)
+
+    session = transport_params.get('session')
+    if session is not None and (uri['access_id'] or uri['access_secret']):
+        logger.warning(
+            'ignoring credentials parsed from URL because they conflict with '
+            'transport_params.session. Set transport_params.session to None '
+            'to suppress this warning.'
+        )
+        uri.update(access_id=None, access_secret=None)
+    elif (uri['access_id'] and uri['access_secret']):
+        transport_params['session'] = boto3.Session(
+            aws_access_key_id=uri['access_id'],
+            aws_secret_access_key=uri['access_secret'],
+        )
+        uri.update(access_id=None, access_secret=None)
+
+    if uri['host'] != DEFAULT_HOST:
+        endpoint_url = 'https://%(host)s:%(port)d' % uri
+        _override_endpoint_url(transport_params, endpoint_url)
+
+    return uri, transport_params
+
+
+def _override_endpoint_url(transport_params, url):
+    try:
+        resource_kwargs = transport_params['resource_kwargs']
+    except KeyError:
+        resource_kwargs = transport_params['resource_kwargs'] = {}
+
+    if resource_kwargs.get('endpoint_url'):
+        logger.warning(
+            'ignoring endpoint_url parsed from URL because it conflicts '
+            'with transport_params.resource_kwargs.endpoint_url. '
+        )
+    else:
+        resource_kwargs.update(endpoint_url=url)
+
+
+def open_uri(uri, mode, transport_params):
+    parsed_uri = parse_uri(uri)
+    parsed_uri, transport_params = _consolidate_params(parsed_uri, transport_params)
+    kwargs = smart_open.utils.check_kwargs(open, transport_params)
+    return open(parsed_uri['bucket_id'], parsed_uri['key_id'], mode, **kwargs)
 
 
 def open(
@@ -94,6 +183,7 @@ def open(
         multipart_upload=True,
         singlepart_upload_kwargs=None,
         object_kwargs=None,
+        defer_seek=False,
         ):
     """Open an S3 object for reading or writing.
 
@@ -132,16 +222,20 @@ def open(
     object_kwargs: dict, optional
         Additional parameters to pass to boto3's object.get function.
         Used during reading only.
-
+    defer_seek: boolean, optional
+        Default: `False`
+        If set to `True` on a file opened for reading, GetObject will not be
+        called until the first seek() or read().
+        Avoids redundant API queries when seeking before reading.
     """
     logger.debug('%r', locals())
-    if mode not in MODES:
-        raise NotImplementedError('bad mode: %r expected one of %r' % (mode, MODES))
+    if mode not in constants.BINARY_MODES:
+        raise NotImplementedError('bad mode: %r expected one of %r' % (mode, constants.BINARY_MODES))
 
-    if (mode == WRITE_BINARY) and (version_id is not None):
+    if (mode == constants.WRITE_BINARY) and (version_id is not None):
         raise ValueError("version_id must be None when writing")
 
-    if mode == READ_BINARY:
+    if mode == constants.READ_BINARY:
         fileobj = Reader(
             bucket_id,
             key_id,
@@ -150,8 +244,9 @@ def open(
             session=session,
             resource_kwargs=resource_kwargs,
             object_kwargs=object_kwargs,
+            defer_seek=defer_seek,
         )
-    elif mode == WRITE_BINARY:
+    elif mode == constants.WRITE_BINARY:
         if multipart_upload:
             fileobj = MultipartWriter(
                 bucket_id,
@@ -171,6 +266,8 @@ def open(
             )
     else:
         assert False, 'unexpected mode: %r' % mode
+
+    fileobj.name = key_id
     return fileobj
 
 
@@ -180,11 +277,21 @@ def _get(s3_object, version=None, **kwargs):
     try:
         return s3_object.get(**kwargs)
     except botocore.client.ClientError as error:
-        raise IOError(
+        wrapped_error = IOError(
             'unable to access bucket: %r key: %r version: %r error: %s' % (
                 s3_object.bucket_name, s3_object.key, version, error
             )
         )
+        wrapped_error.backend_error = error
+        raise wrapped_error from error
+
+
+def _unwrap_ioerror(ioe):
+    """Given an IOError from _get, return the 'Error' dictionary from boto."""
+    try:
+        return ioe.backend_error.response['Error']
+    except (AttributeError, KeyError):
+        return None
 
 
 class _SeekableRawReader(object):
@@ -193,19 +300,26 @@ class _SeekableRawReader(object):
     This class is internal to the S3 submodule.
     """
 
-    def __init__(self, s3_object, content_length, version_id=None, object_kwargs=None):
+    def __init__(self, s3_object, version_id=None, object_kwargs=None):
         self._object = s3_object
-        self._content_length = content_length
+        self._content_length = None
         self._version_id = version_id
         self._position = 0
         self._body = None
         self._object_kwargs = object_kwargs if object_kwargs else {}
 
-    def seek(self, position):
-        """Seek to the specified position (byte offset) in the S3 key.
+    def seek(self, offset, whence=constants.WHENCE_START):
+        """Seek to the specified position.
 
-        :param int position: The byte offset from the beginning of the key.
+        :param int offset: The offset in bytes.
+        :param int whence: Where the offset is from.
+
+        :returns: the position after seeking.
+        :rtype: int
         """
+        if whence not in constants.WHENCE_CHOICES:
+            raise ValueError('invalid whence, expected one of %r' % constants.WHENCE_CHOICES)
+
         #
         # Close old body explicitly.
         # When first seek() after __init__(), self._body is not exist.
@@ -213,27 +327,82 @@ class _SeekableRawReader(object):
         if self._body is not None:
             self._body.close()
         self._body = None
-        self._position = position
 
-    def _load_body(self):
-        """Build a continuous connection with the remote peer starts from the current postion.
-        """
-        range_string = make_range_string(self._position)
-        logger.debug('content_length: %r range_string: %r', self._content_length, range_string)
-
-        if self._position == self._content_length == 0 or self._position == self._content_length:
-            #
-            # When reading, we can't seek to the first byte of an empty file.
-            # Similarly, we can't seek past the last byte.  Do nothing here.
-            #
-            self._body = io.BytesIO()
+        start = None
+        stop = None
+        if whence == constants.WHENCE_START:
+            start = max(0, offset)
+        elif whence == constants.WHENCE_CURRENT:
+            start = max(0, offset + self._position)
         else:
-            self._body = _get(
+            stop = max(0, -offset)
+
+        #
+        # If we can figure out that we've read past the EOF, then we can save
+        # an extra API call.
+        #
+        if self._content_length is None:
+            reached_eof = False
+        elif start is not None and start >= self._content_length:
+            reached_eof = True
+        elif stop == 0:
+            reached_eof = True
+        else:
+            reached_eof = False
+
+        if reached_eof:
+            self._body = io.BytesIO()
+            self._position = self._content_length
+        else:
+            self._open_body(start, stop)
+
+        return self._position
+
+    def _open_body(self, start=None, stop=None):
+        """Open a connection to download the specified range of bytes. Store
+        the open file handle in self._body.
+
+        If no range is specified, start defaults to self._position.
+        start and stop follow the semantics of the http range header,
+        so a stop without a start will read bytes beginning at stop.
+
+        As a side effect, set self._content_length. Set self._position
+        to self._content_length if start is past end of file.
+        """
+        if start is None and stop is None:
+            start = self._position
+        range_string = smart_open.utils.make_range_string(start, stop)
+        logger.debug('range_string: %r', range_string)
+
+        try:
+            # Optimistically try to fetch the requested content range.
+            response = _get(
                 self._object,
                 version=self._version_id,
                 Range=range_string,
                 **self._object_kwargs
-            )['Body']
+            )
+        except IOError as ioe:
+            # Handle requested content range exceeding content size.
+            error_response = _unwrap_ioerror(ioe)
+            if error_response is None or error_response.get('Message') != _OUT_OF_RANGE:
+                raise
+            try:
+                self._position = self._content_length = int(error_response['ActualObjectSize'])
+            except KeyError:
+                # This shouldn't happen with real S3, but moto lacks ActualObjectSize.
+                # Reported at https://github.com/spulec/moto/issues/2981
+                self._position = self._content_length = _get(
+                    self._object,
+                    version=self._version_id,
+                    **self._object_kwargs,
+                )['ContentLength']
+            self._body = io.BytesIO()
+        else:
+            units, start, stop, length = smart_open.utils.parse_content_range(response['ContentRange'])
+            self._content_length = length
+            self._position = start
+            self._body = response['Body']
 
     def _read_from_body(self, size=-1):
         if size == -1:
@@ -244,17 +413,17 @@ class _SeekableRawReader(object):
 
     def read(self, size=-1):
         """Read from the continuous connection with the remote peer."""
+        if self._body is None:
+            # This is necessary for the very first read() after __init__().
+            self._open_body()
         if self._position >= self._content_length:
             return b''
-        if self._body is None:
-            # When the first read() after __init__() or seek(), self._body is not exist.
-            self._load_body()
 
         try:
             binary = self._read_from_body(size)
         except botocore.exceptions.IncompleteReadError:
             # The underlying connection of the self._body was closed by the remote peer.
-            self._load_body()
+            self._open_body()
             binary = self._read_from_body(size)
         self._position += len(binary)
         return binary
@@ -266,8 +435,8 @@ class Reader(io.BufferedIOBase):
     Implements the io.BufferedIOBase interface of the standard library."""
 
     def __init__(self, bucket, key, version_id=None, buffer_size=DEFAULT_BUFFER_SIZE,
-                 line_terminator=BINARY_NEWLINE, session=None, resource_kwargs=None,
-                 object_kwargs=None):
+                 line_terminator=constants.BINARY_NEWLINE, session=None, resource_kwargs=None,
+                 object_kwargs=None, defer_seek=False):
 
         self._buffer_size = buffer_size
 
@@ -285,15 +454,9 @@ class Reader(io.BufferedIOBase):
         s3 = session.resource('s3', **resource_kwargs)
         self._object = s3.Object(bucket, key)
         self._version_id = version_id
-        self._content_length = _get(
-            self._object,
-            version=self._version_id,
-            **self._object_kwargs
-        )['ContentLength']
 
         self._raw_reader = _SeekableRawReader(
             self._object,
-            self._content_length,
             self._version_id,
             self._object_kwargs,
         )
@@ -307,6 +470,9 @@ class Reader(io.BufferedIOBase):
         #
         self.raw = None
 
+        if not defer_seek:
+            self.seek(0)
+
     #
     # io.BufferedIOBase methods.
     #
@@ -325,9 +491,10 @@ class Reader(io.BufferedIOBase):
         if size == 0:
             return b''
         elif size < 0:
-            from_buf = self._read_from_buffer()
-            self._current_pos = self._content_length
-            return from_buf + self._raw_reader.read()
+            # call read() before setting _current_pos to make sure _content_length is set
+            out = self._read_from_buffer() + self._raw_reader.read()
+            self._current_pos = self._raw_reader._content_length
+            return out
 
         #
         # Return unused data first
@@ -384,7 +551,7 @@ class Reader(io.BufferedIOBase):
         We offer only seek support, and no truncate support."""
         return True
 
-    def seek(self, offset, whence=START):
+    def seek(self, offset, whence=constants.WHENCE_START):
         """Seek to the specified position.
 
         :param int offset: The offset in bytes.
@@ -392,22 +559,18 @@ class Reader(io.BufferedIOBase):
 
         Returns the position after seeking."""
         logger.debug('seeking to offset: %r whence: %r', offset, whence)
-        if whence not in WHENCE_CHOICES:
-            raise ValueError('invalid whence, expected one of %r' % WHENCE_CHOICES)
 
-        if whence == START:
-            new_position = offset
-        elif whence == CURRENT:
-            new_position = self._current_pos + offset
-        else:
-            new_position = self._content_length + offset
-        new_position = clamp(new_position, 0, self._content_length)
-        self._current_pos = new_position
-        self._raw_reader.seek(new_position)
+        # Convert relative offset to absolute, since self._raw_reader
+        # doesn't know our current position.
+        if whence == constants.WHENCE_CURRENT:
+            whence = constants.WHENCE_START
+            offset += self._current_pos
+
+        self._current_pos = self._raw_reader.seek(offset, whence)
         logger.debug('new_position: %r', self._current_pos)
 
         self._buffer.empty()
-        self._eof = self._current_pos == self._content_length
+        self._eof = self._current_pos == self._raw_reader._content_length
         return self._current_pos
 
     def tell(self):
@@ -436,7 +599,10 @@ class Reader(io.BufferedIOBase):
 
         """
         s3 = self._session.resource('s3', **self._resource_kwargs)
-        return s3.Object(self._object.bucket_name, self._object.key)
+        if self._version_id is not None:
+            return s3.Object(self._object.bucket_name, self._object.key).Version(self._version_id)
+        else:
+            return s3.Object(self._object.bucket_name, self._object.key)
 
     #
     # Internal methods.
@@ -524,7 +690,7 @@ multipart upload may fail")
                 'the bucket %r does not exist, or is forbidden for access (%r)' % (
                     bucket, error
                 )
-            )
+            ) from error
 
         self._buf = io.BytesIO()
         self._total_bytes = 0
@@ -704,8 +870,8 @@ class SinglepartWriter(io.BufferedIOBase):
         try:
             self._object = s3.Object(bucket, key)
             s3.meta.client.head_bucket(Bucket=bucket)
-        except botocore.client.ClientError:
-            raise ValueError('the bucket %r does not exist, or is forbidden for access' % bucket)
+        except botocore.client.ClientError as e:
+            raise ValueError('the bucket %r does not exist, or is forbidden for access' % bucket) from e
 
         self._buf = io.BytesIO()
         self._total_bytes = 0
@@ -729,9 +895,9 @@ class SinglepartWriter(io.BufferedIOBase):
 
         try:
             self._object.put(Body=self._buf, **self._upload_kwargs)
-        except botocore.client.ClientError:
+        except botocore.client.ClientError as e:
             raise ValueError(
-                'the bucket %r does not exist, or is forbidden for access' % self._object.bucket_name)
+                'the bucket %r does not exist, or is forbidden for access' % self._object.bucket_name) from e
 
         logger.debug("direct upload finished")
         self._buf = None
@@ -801,7 +967,9 @@ def _retry_if_failed(
         partial,
         attempts=_UPLOAD_ATTEMPTS,
         sleep_seconds=_SLEEP_SECONDS,
-        exceptions=(botocore.exceptions.EndpointConnectionError, )):
+        exceptions=None):
+    if exceptions is None:
+        exceptions = (botocore.exceptions.EndpointConnectionError, )
     for attempt in range(attempts):
         try:
             return partial()
@@ -911,7 +1079,7 @@ def iter_bucket(
         retries=retries,
         **session_kwargs)
 
-    with _create_process_pool(processes=workers) as pool:
+    with smart_open.concurrency.create_pool(processes=workers) as pool:
         result_iterator = pool.imap_unordered(download_key, key_iterator)
         for key_no, (key, content) in enumerate(result_iterator):
             if True or key_no % 1000 == 0:
@@ -994,41 +1162,3 @@ def _download_fileobj(bucket, key_name):
     buf = io.BytesIO()
     bucket.download_fileobj(key_name, buf)
     return buf.getvalue()
-
-
-class DummyPool(object):
-    """A class that mimics multiprocessing.pool.Pool for our purposes."""
-    def imap_unordered(self, function, items):
-        return six.moves.map(function, items)
-
-    def terminate(self):
-        pass
-
-
-class ConcurrentFuturesPool(object):
-    """A class that mimics multiprocessing.pool.Pool but uses concurrent futures instead of processes."""
-    def __init__(self, max_workers):
-        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers)
-
-    def imap_unordered(self, function, items):
-        futures = [self.executor.submit(function, item) for item in items]
-        for future in concurrent.futures.as_completed(futures):
-            yield future.result()
-
-    def terminate(self):
-        self.executor.shutdown(wait=True)
-
-
-@contextlib.contextmanager
-def _create_process_pool(processes=1):
-    if _MULTIPROCESSING and processes:
-        logger.info("creating multiprocessing pool with %i workers", processes)
-        pool = multiprocessing.pool.Pool(processes=processes)
-    elif _CONCURRENT_FUTURES and processes:
-        logger.info("creating concurrent futures pool with %i workers", processes)
-        pool = ConcurrentFuturesPool(max_workers=processes)
-    else:
-        logger.info("creating dummy pool")
-        pool = DummyPool()
-    yield pool
-    pool.terminate()
