@@ -957,18 +957,18 @@ class SmartOpenReadTest(unittest.TestCase):
 class SmartOpenS3KwargsTest(unittest.TestCase):
     @mock.patch('boto3.Session')
     def test_no_kwargs(self, mock_session):
-        smart_open.smart_open('s3://mybucket/mykey')
+        smart_open.smart_open('s3://mybucket/mykey', defer_seek=True)
         mock_session.return_value.resource.assert_called_with('s3')
 
     @mock.patch('boto3.Session')
     def test_credentials(self, mock_session):
-        smart_open.smart_open('s3://access_id:access_secret@mybucket/mykey')
+        smart_open.smart_open('s3://access_id:access_secret@mybucket/mykey', defer_seek=True)
         mock_session.assert_called_with(aws_access_key_id='access_id', aws_secret_access_key='access_secret')
         mock_session.return_value.resource.assert_called_with('s3')
 
     @mock.patch('boto3.Session')
     def test_host(self, mock_session):
-        transport_params = {'resource_kwargs': {'endpoint_url': 'http://aa.domain.com'}}
+        transport_params = {'resource_kwargs': {'endpoint_url': 'http://aa.domain.com'}, 'defer_seek': True}
         smart_open.open("s3://access_id:access_secret@mybucket/mykey", transport_params=transport_params)
         mock_session.assert_called_with(
             aws_access_key_id='access_id',
@@ -1008,7 +1008,7 @@ class SmartOpenS3KwargsTest(unittest.TestCase):
         session = boto3.Session()
         session.resource = mock.MagicMock()
 
-        smart_open.open('s3://bucket/key', transport_params={'session': session})
+        smart_open.open('s3://bucket/key', transport_params={'session': session, 'defer_seek': True})
         session.resource.assert_called_with('s3')
 
     def test_session_write_mode(self):
